// markov.c
// Markovi masin 1998-04-12 programmeerimisvistluse 3. lesande testimiseks
// Autor: Ahto Truu

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define MAX_KSK 1024
#define MAX_REG 10240
#define MAX_AEG 200000

typedef struct {
	int arv;
	char **vas, **par;
} programm;

int loe_programm (char *str, programm *prg);
int lisa_kask (programm *prg, char *vas, char *par);
int loe_andmed (char *str, char *reg);
int taida_programm (programm *prg, char *reg);
int asenda_string (char *reg, char *pos, char *vas, char *par);
int kirjuta_andmed (char *str, char *reg);

int main (int argc, char **argv)
{
	programm p;
	char r[MAX_REG + 1];

	if (argc != 4) {
		fprintf (stderr, "Sntaks on: %s programm sisend vljund\n", argv[0]);
		return 0;
	}

	if (loe_programm (argv[1], &p))
		if (loe_andmed (argv[2], r))
			if (taida_programm (&p, r))
				if (kirjuta_andmed (argv[3], r))
					return 1;

	return 0;
}

int loe_programm (char *str, programm *prg)
{
	FILE *f;
	char b[MAX_KSK + 1];
	int c;

	prg->arv = 0; prg->vas = NULL; prg->par = NULL;

	if ((f = fopen (str, "rt")) == NULL) {
		fprintf (stderr, "Viga faili %s avamisel\n", str);
		return 0;
	}

	while ((c = fgetc (f)) != EOF)
		if (c == '(') {
			int i = 0, x;
			while (i < MAX_KSK && (c = fgetc (f)) != EOF && c != ',')
				b[i++] = c;
			if (i == MAX_KSK) {
				fprintf (stderr, "Liiga pikk ksk\n");
				return 0;
			}
			if (c == EOF) {
				fprintf (stderr, "Lpetamata ksk faili lpus\n");
				return 0;
			}
			b[i++] = 0; x = i;
			while (i < MAX_KSK && (c = fgetc (f)) != EOF && c != ')')
				b[i++] = c;
			if (i == MAX_KSK) {
				fprintf (stderr, "Liiga pikk ksk\n");
				return 0;
			}
			if (c == EOF) {
				fprintf (stderr, "Lpetamata ksk faili lpus\n");
				return 0;
			}
			b[i] = 0;
			if (!lisa_kask (prg, b, b + x))
				return 0;
		}

	fclose (f);
	return 1;
}

int lisa_kask (programm *prg, char *vas, char *par)
{
	if ((prg->arv & (prg->arv + 1)) == 0) {
		int n = 2 * prg->arv + !prg->arv;
		if ((prg->vas = realloc (prg->vas, n * sizeof (char *))) == NULL) {
			fprintf (stderr, "Mlu otsas\n");
			return 0;
		}
		if ((prg->par = realloc (prg->par, n * sizeof (char *))) == NULL) {
			fprintf (stderr, "Mlu otsas\n");
			return 0;
		}
	}

	if ((prg->vas[prg->arv] = strdup (vas)) == NULL) {
		fprintf (stderr, "Mlu otsas\n");
		return 0;
	}
	if ((prg->par[prg->arv] = strdup (par)) == NULL) {
		fprintf (stderr, "Mlu otsas\n");
		return 0;
	}

	return ++prg->arv;
}

int loe_andmed (char *str, char *reg)
{
	FILE *f;
	int l;

	if ((f = fopen (str, "rt")) == NULL) {
		fprintf (stderr, "Viga faili %s avamisel\n", str);
		return 0;
	}

	if (fgets (reg, MAX_REG + 1, f) == NULL) {
		fprintf (stderr, "Viga faili %s lugemisel\n", str);
		return 0;
	}

	if ((l = strlen (reg)) > 0 && reg[l - 1] == '\n')
		reg[l - 1] = 0;

	fclose (f);
	return 1;
}

int taida_programm (programm *prg, char *reg)
{
	long l = 0;
	int i = 0, c;
	char *p;

	while (l < MAX_AEG && i < prg->arv) {
		if ((p = strstr (reg, prg->vas[i])) == NULL)
			++i;
		else
			if (asenda_string (reg, p, prg->vas[i], prg->par[i]))
				i = 0;
			else
				return 0;
		if (++l % 1000 == 0)
			fprintf (stderr, "          \r%ld\r", l);
	}
	fprintf (stderr, "          \r");

	if (l == MAX_AEG) {
		fprintf (stderr, "Ajapiir letatatud\n");
		return 0;
	}

	return 1;
}

int asenda_string (char *reg, char *pos, char *vas, char *par)
{
	int rp = strlen (reg), vp = strlen (vas), pp = strlen (par);

	if (rp - vp + pp > MAX_REG) {
		fprintf (stderr, "Registri letitumine\n");
		return 0;
	}

	if (vp != pp)
		memmove (pos + pp, pos + vp, rp - (pos - reg) - vp + 1);
	memcpy (pos, par, pp);

	return 1;
}

int kirjuta_andmed (char *str, char *reg)
{
	FILE *f;

	if ((f = fopen (str, "wt")) == NULL) {
		fprintf (stderr, "Viga faili %s avamisel\n", str);
		return 0;
	}

	if (fputs (reg, f) == EOF || fputs ("\n", f) == EOF) {
		fprintf (stderr, "Viga faili %s kirjutamisel\n", str);
		return 0;
	}

	fclose (f);
	return 1;
}